<?php

namespace FiloBlu\Esb\Model;

use Exception;
use FiloBlu\Esb\Api\Data\ConsumerInterface;
use FiloBlu\Esb\Api\Data\QueueItemInterface;
use FiloBlu\Esb\Api\Data\StatusInterface;
use FiloBlu\Esb\Api\DispatcherInterface;
use FiloBlu\Esb\Api\QueueItemRepositoryInterface;
use FiloBlu\Esb\Api\ConsumerResolverInterface;
use FiloBlu\Esb\Api\RetryStrategyInterface;
use Magento\Framework\Api\SearchCriteriaBuilder;
use Magento\Framework\Api\SearchCriteriaBuilderFactory;
use Psr\Log\LoggerInterface;
use Throwable;

/**
 * Class Dispatcher
 * @package FiloBlu\Esb\Model
 */
class Dispatcher implements DispatcherInterface
{
    /**
     * @var QueueItemRepositoryInterface
     */
    protected $queueItemRepository;
    /**
     * @var SearchCriteriaBuilderFactory
     */
    protected $searchCriteriaBuilderFactory;
    /**
     * @var LoggerInterface
     */
    protected $logger;
    /**
     * @var ConsumerResolverInterface
     */
    private $consumerResolver;
    /**
     * @var RetryStrategyInterface
     */
    private $retryStrategy;

    /**
     * Dispatcher constructor.
     * @param LoggerInterface $logger
     * @param ConsumerResolverInterface $consumerResolver
     * @param QueueItemRepositoryInterface $queueItemRepository
     * @param RetryStrategyInterface $retryStrategy
     * @param SearchCriteriaBuilderFactory $searchCriteriaBuilderFactory
     */
    public function __construct(LoggerInterface $logger,
                                ConsumerResolverInterface $consumerResolver,
                                QueueItemRepositoryInterface $queueItemRepository,
                                RetryStrategyInterface $retryStrategy,
                                SearchCriteriaBuilderFactory $searchCriteriaBuilderFactory)
    {
        $this->searchCriteriaBuilderFactory = $searchCriteriaBuilderFactory;
        $this->queueItemRepository = $queueItemRepository;
        $this->logger = $logger;
        $this->consumerResolver = $consumerResolver;
        $this->retryStrategy = $retryStrategy;
    }

    /**
     *
     */
    public function dispatch()
    {
        /** @var SearchCriteriaBuilder $builder */
        $builder = $this->searchCriteriaBuilderFactory->create();
        $builder->setSortOrders(['priority' => 'asc']);

        /** @var QueueItemInterface[] $items */
        $items = $this->queueItemRepository->getList($builder->create());

        foreach ($items as $item) {
            $status = null;
            try {
                /** @var ConsumerInterface $consumer */
                $consumer = $this->consumerResolver->get($item->getEvent());
                $status = $consumer->consume($item->getMessage());
            } catch (Exception $exception) {
                $status = $this->handleFailure($item, $exception);
            } catch (Throwable $throwable) {
                $status = $this->handleFailure($item, $throwable);
            } finally {
                $item->setStatus($status);
                $this->queueItemRepository->save($item);
            }
        }
    }

    /**
     * @param QueueItemInterface $item
     * @param Throwable $exception
     * @return StatusInterface
     */
    protected function handleFailure(QueueItemInterface $item, Throwable $exception): StatusInterface
    {
        $id = $item->getId();
        $status = $item->getStatus();
        $status->setCode('error');
        $message = $exception->getMessage();

        if ($this->retryStrategy->reschedule($item)) {
            $message = "Rescheduled item with id {$id} because exception : {$message}";
        } else {
            $message = "Item with {$id} will not be scheduled anymore : {$message}";
        }

        $this->logger->error($message, ['exception' => $exception]);

        return $status;
    }
}
